"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parallelLimit = void 0;
const queue_1 = require("@zodash/queue");
const event_1 = require("@zodash/event");
const next_tick_1 = require("@zodash/next-tick");
const nextTick = (fn) => next_tick_1.nextTick(fn, 200);
// const poll = (queue: Queue<ITask<any>>) => {
//   if (queue.isEmpty()) {
//     nextTick(poll, queue);
//   }
//   const task = queue.dequeue();
//   task((err, result) => {
//     nextTick(poll, queue);
//   });
// };
// const poll = (generators: any) => {
//   if (generators.done()) {
//     nextTick(poll, generators);
//   }
//   const task = generators.next();
//   task((err, result) => {
//     nextTick(poll, generators);
//   });
// };
function toPromise(fn) {
    if (fn.length === 1) {
        return () => new Promise((resolve, reject) => fn.call(null, (error, result) => {
            if (error) {
                return reject(error);
            }
            return resolve(result);
        }));
    }
    return fn;
}
function parallelLimit(tasks, limit, cb) {
    // const store = new Cache<number, { id: number, index: number, task: ITask<R> }>(Infinity);
    const emitter = new event_1.Event();
    const running = new queue_1.Queue(limit);
    const pending = new queue_1.Queue();
    const results = [];
    function done() {
        if (cb) {
            cb(results);
            cb = null;
        }
        else {
            emitter.emit('done', results);
        }
    }
    function next() {
        // Full, Go Next Tick
        if (running.isFull()) {
            return nextTick(next);
        }
        // no rest tasks
        if (pending.isEmpty()) {
            // no running tasks
            if (running.isEmpty()) {
                return nextTick(done);
            }
            return;
        }
        const value = pending.dequeue();
        running.enqueue(value);
        const { task, index } = value;
        toPromise(task)()
            .then((result) => {
            results[index] = result;
        })
            .catch((error) => {
            results[index] = error;
        })
            .finally(() => {
            running.dequeue();
            // give err and result to next
            nextTick(next);
        });
    }
    function coordinator() {
        if (running.isFull()) {
            return nextTick(coordinator);
        }
        const restRunningSize = running.restSize();
        for (let i = 0; i < restRunningSize; ++i) {
            nextTick(next);
        }
    }
    function run() {
        // empty
        if (!tasks.length) {
            return done();
        }
        tasks.forEach((task, index) => {
            const id = index;
            pending.enqueue({
                id,
                task,
                index,
            });
        });
        return nextTick(coordinator);
    }
    // setInterval(() => {
    //   console.log('current running worker:', running.size());
    // }, 300);
    run();
    if (!cb) {
        return new Promise((resolve) => {
            emitter.on('done', resolve);
        });
    }
}
exports.parallelLimit = parallelLimit;
